@@ -1,4 +1,5 @@
 import dataclasses
+import re
 import typing as tp
 
 import jax
@@ -13,7 +14,7 @@ key = jax.random.PRNGKey
 _pymap = map
 _pyfilter = filter
 
-OpaqueIsEqual = tp.Optional[tp.Callable[["Opaque", tp.Any], bool]]
+OpaquePredicate = tp.Callable[["Opaque", tp.Any], bool]
 
 
 @tpe.runtime_checkable
@@ -23,16 +24,16 @@ class ArrayLike(tpe.Protocol):
 
 
 class Opaque(tp.Generic[A]):
-    def __init__(self, value: A, opaque_is_equal: tp.Optional[OpaqueIsEqual]):
+    def __init__(self, value: A, predicate: tp.Optional[OpaquePredicate]):
         self.value = value
-        self.opaque_is_equal = opaque_is_equal
+        self.predicate = predicate
 
     def __repr__(self):
         return f"Hidden({self.value})"
 
     def __eq__(self, other):
-        if self.opaque_is_equal is not None:
-            return self.opaque_is_equal(self, other)
+        if self.predicate is not None:
+            return self.predicate(self, other)
         else:
             # if both are Opaque and their values are of the same type
             if isinstance(other, Opaque) and type(self.value) == type(other.value):
@@ -51,17 +52,16 @@ class Opaque(tp.Generic[A]):
 
 
 def field(
-    default=dataclasses.MISSING,
+    default: tp.Any = dataclasses.MISSING,
     *,
     node: bool,
     kind: type = type(None),
-    default_factory=dataclasses.MISSING,
+    default_factory: tp.Optional[tp.Callable[[], tp.Any]] = None,
     init: bool = True,
     repr: bool = True,
     hash: tp.Optional[bool] = None,
     compare: bool = True,
-    opaque: bool = False,
-    opaque_is_equal: tp.Optional[tp.Callable[[Opaque, tp.Any], bool]] = None,
+    opaque: tp.Union[bool, OpaquePredicate] = False,
 ) -> tp.Any:
 
     return dataclasses.field(
@@ -70,9 +70,10 @@ def field(
             "node": node,
             "kind": kind,
             "opaque": opaque,
-            "opaque_is_equal": opaque_is_equal,
         },
-        default_factory=default_factory,
+        default_factory=default_factory
+        if default_factory is not None
+        else dataclasses.MISSING,
         init=init,
         repr=repr,
         hash=hash,
@@ -84,13 +85,12 @@ def node(
     default=dataclasses.MISSING,
     *,
     kind: type = type(None),
-    default_factory=dataclasses.MISSING,
+    default_factory: tp.Optional[tp.Callable[[], tp.Any]] = None,
     init: bool = True,
     repr: bool = True,
     hash: tp.Optional[bool] = None,
     compare: bool = True,
-    opaque: bool = False,
-    opaque_is_equal: tp.Optional[tp.Callable[[Opaque, tp.Any], bool]] = None,
+    opaque: tp.Union[bool, OpaquePredicate] = False,
 ) -> tp.Any:
     return field(
         default=default,
@@ -102,7 +102,6 @@ def node(
         hash=hash,
         compare=compare,
         opaque=opaque,
-        opaque_is_equal=opaque_is_equal,
     )
 
 
@@ -110,13 +109,12 @@ def static(
     default=dataclasses.MISSING,
     *,
     kind: type = type(None),
-    default_factory=dataclasses.MISSING,
+    default_factory: tp.Optional[tp.Callable[[], tp.Any]] = None,
     init: bool = True,
     repr: bool = True,
     hash: tp.Optional[bool] = None,
     compare: bool = True,
-    opaque: bool = False,
-    opaque_is_equal: tp.Optional[tp.Callable[[Opaque, tp.Any], bool]] = None,
+    opaque: tp.Union[bool, OpaquePredicate] = False,
 ) -> tp.Any:
     return field(
         default,
@@ -128,7 +126,6 @@ def static(
         hash=hash,
         compare=compare,
         opaque=opaque,
-        opaque_is_equal=opaque_is_equal,
     )
 
 
@@ -160,3 +157,65 @@ def _all_types_unfiltered(t: type) -> tp.Iterable[type]:
 
         for arg in t.__args__:
             yield from _all_types_unfiltered(arg)
+
+
+def _unique_name(
+    names: tp.Set[str],
+    name: str,
+):
+
+    if name in names:
+
+        match = re.match(r"(.*?)(\d*)$", name)
+        assert match is not None
+
+        name = match[1]
+        num_part = match[2]
+
+        i = int(num_part) if num_part else 2
+        str_template = f"{{name}}{{i:0{len(num_part)}}}"
+
+        while str_template.format(name=name, i=i) in names:
+            i += 1
+
+        name = str_template.format(name=name, i=i)
+
+    names.add(name)
+    return name
+
+
+def _unique_names(
+    names: tp.Iterable[str],
+    *,
+    existing_names: tp.Optional[tp.Set[str]] = None,
+) -> tp.Iterable[str]:
+    if existing_names is None:
+        existing_names = set()
+
+    for name in names:
+        yield _unique_name(existing_names, name)
+
+
+def _lower_snake_case(s: str) -> str:
+    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
+    parts = s.split("_")
+    output_parts = []
+
+    for i in range(len(parts)):
+        if i == 0 or len(parts[i - 1]) > 1:
+            output_parts.append(parts[i])
+        else:
+            output_parts[-1] += parts[i]
+
+    return "_".join(output_parts)
+
+
+def _get_name(obj) -> str:
+    if hasattr(obj, "name") and obj.name:
+        return obj.name
+    elif hasattr(obj, "__name__") and obj.__name__:
+        return obj.__name__
+    elif hasattr(obj, "__class__") and obj.__class__.__name__:
+        return _lower_snake_case(obj.__class__.__name__)
+    else:
+        raise ValueError(f"Could not get name for: {obj}")
